#include "raylib.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <math.h>

#define SIZE_OF_HEX 20

struct celluleDemineur{
    int line, column;
    bool hasMine;
    int nearbyMine;
    Vector2 pos;
    bool isPressed;
    bool isFlaged;
    bool isNeareast;
    Color couleurCase;
};

typedef struct celluleDemineur* CelluleDemineur;

bool containsMine(int line, int col, CelluleDemineur** matriceMines){
    return (matriceMines[line][col]->hasMine == true);
}

void afficherGrilleConsole(int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines){
    for(int i=0; i<nbLigneMatrice; i++){
        for(int j=0; j<nbColonneMatrice; j++){
            printf("M%d:N%d - ", matriceMines[i][j]->hasMine, matriceMines[i][j]->nearbyMine);
        }
        printf("\n");
    }
}

CelluleDemineur* recupererVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* tabVoisins = calloc(6, sizeof(struct celluleDemineur));
    for(int i=0; i<6; i++){
        tabVoisins[i] = NULL;
    }
    
    bool linePrev = (line-1 >= 0), colPrev = (column-1 >= 0);
    bool lineNext = (line+1 < nbLigneMatrice), colNext = (column+1 < nbColonneMatrice);
    
    if(colPrev){tabVoisins[0] = matriceMines[line][column-1];}
    if(colNext){tabVoisins[1] = matriceMines[line][column+1];}
    
    if(linePrev){tabVoisins[2] = matriceMines[line-1][column];}
    if(lineNext){tabVoisins[3] = matriceMines[line+1][column];}
    
    if(line%2 == 0){
        if(linePrev && colNext){tabVoisins[4] = matriceMines[line-1][column+1];}
        if(lineNext && colNext){tabVoisins[5] = matriceMines[line+1][column+1];}
    }else{
        if(linePrev && colPrev){tabVoisins[4] = matriceMines[line-1][column-1];}
        if(lineNext && colPrev){tabVoisins[5] = matriceMines[line+1][column-1];}
    }
    
    return tabVoisins;
}

void incrementationVoisinMine(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* voisins = recupererVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
    
    for(int i=0; i<6; i++){
        if(voisins[i] != NULL){
            voisins[i]->nearbyMine++;
        }
    }
    free(voisins);
}

void revelerVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* voisins = recupererVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
    
    for(int i=0; i<6; i++){
        if(voisins[i] != NULL){
            revelerCase(matriceMines, voisins[i]->line, voisins[i]->column, nbLigneMatrice, nbColonneMatrice);
        }
    }
    free(voisins);
}

void generationMines(int nbMines, int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines){
    for(int i=0; i<nbMines; i++){
        int lineRand = rand()%nbLigneMatrice;
        int colRand = rand()%nbColonneMatrice;
        
        if(!containsMine(lineRand, colRand, matriceMines)){
            matriceMines[lineRand][colRand]->hasMine = true;
            matriceMines[lineRand][colRand]->nearbyMine = -9;
            incrementationVoisinMine(lineRand, colRand, matriceMines, nbLigneMatrice, nbColonneMatrice);
        }
    }
}

void initGrille(int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines){
    for(int i=0; i<nbLigneMatrice; i++){
        for(int j=0; j<nbColonneMatrice; j++){
            matriceMines[i][j] = malloc(sizeof(struct celluleDemineur));
            matriceMines[i][j]->hasMine = false;
            matriceMines[i][j]->isPressed = false;
            matriceMines[i][j]->isFlaged = false;
            matriceMines[i][j]->isNeareast = false;
            matriceMines[i][j]->nearbyMine = 0;
            matriceMines[i][j]->line = i;
            matriceMines[i][j]->column = j;
            matriceMines[i][j]->couleurCase = DARKGRAY;
            
            if(i%2 == 0){
                matriceMines[i][j]->pos = (Vector2){(1.5*j+1.75)*SIZE_OF_HEX, (1.5*i+1)*SIZE_OF_HEX};
            }else{
                matriceMines[i][j]->pos = (Vector2){(1.5*j+1)*SIZE_OF_HEX, (1.5*i+1)*SIZE_OF_HEX};
            }
        }
    }
}

void revelerCase(CelluleDemineur** matriceMines, int line, int column, int nbLigneMatrice, int nbColonneMatrice){
    if(!matriceMines[line][column]->isFlaged && !matriceMines[line][column]->isPressed){
        matriceMines[line][column]->isPressed = true;
    
        if(matriceMines[line][column]->nearbyMine == 0){
            matriceMines[line][column]->couleurCase = LIGHTGRAY;
            revelerVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
        }else if(matriceMines[line][column]->nearbyMine < 0){
            matriceMines[line][column]->couleurCase = MAROON;
        }else{
            matriceMines[line][column]->couleurCase = LIGHTGRAY;
        }
    }
}

float getDistanceBetweenPoints(Vector2 a, Vector2 b) {
  return (a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y);
}

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 800;
    const int screenHeight = 600;
    
    srand(time(NULL));
    
    //CREER MATRICE DES MINES
    int nbLigneMatrice = 16, nbColonneMatrice = 21, nbMines = 60;
    CelluleDemineur** matriceMines = malloc(nbLigneMatrice * sizeof(CelluleDemineur *));
    for(int i=0; i<nbColonneMatrice; i++)
        matriceMines[i] = malloc(nbColonneMatrice * sizeof(CelluleDemineur));

    initGrille(nbLigneMatrice, nbColonneMatrice, matriceMines);
    
    //GENERATION DES MINES
    generationMines(nbMines, nbLigneMatrice, nbColonneMatrice, matriceMines);
    
    int remainingMines = nbMines, curCaseX = 0, curCaseY = 0;

    InitWindow(screenWidth, screenHeight, "HexaSweeper");
    
    

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second
    //--------------------------------------------------------------------------------------

    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        // Update
        //----------------------------------------------------------------------------------
        if(IsMouseButtonPressed(MOUSE_LEFT_BUTTON)){
            revelerCase(matriceMines, curCaseX, curCaseY, nbLigneMatrice, nbColonneMatrice);
        }
        
        if(IsMouseButtonPressed(MOUSE_RIGHT_BUTTON)){
            if(!matriceMines[curCaseX][curCaseY]->isPressed){
                matriceMines[curCaseX][curCaseY]->isFlaged = !matriceMines[0][0]->isFlaged;
                if(matriceMines[curCaseX][curCaseY]->isFlaged){
                    matriceMines[curCaseX][curCaseY]->couleurCase = GREEN;
                    remainingMines--;
                }
                else{
                    matriceMines[curCaseX][curCaseY]->couleurCase = DARKGRAY;
                    remainingMines++;
                }
            }
        }
        
        if(IsMouseButtonPressed(MOUSE_MIDDLE_BUTTON)){
            revelerCase(matriceMines, curCaseX, curCaseY, nbLigneMatrice, nbColonneMatrice);
            revelerVoisins(curCaseX, curCaseY, matriceMines, nbLigneMatrice, nbColonneMatrice);
        }
        
        Vector2 mPos = GetMousePosition();
        for(int i=0; i<nbLigneMatrice; i++){
            for(int j=0; j<nbColonneMatrice; j++){
                float distanceCur = getDistanceBetweenPoints(mPos, matriceMines[i][j]->pos);
                if(distanceCur < SIZE_OF_HEX*SIZE_OF_HEX/2){
                    matriceMines[i][j]->isNeareast = true;
                    curCaseX = matriceMines[i][j]->line;
                    curCaseY = matriceMines[i][j]->column;
                }else{
                    matriceMines[i][j]->isNeareast = false;
                }
            }
        }
        
        // Draw
        //----------------------------------------------------------------------------------
        BeginDrawing();

            ClearBackground(RAYWHITE);
            
            for(int i=0; i<nbLigneMatrice; i++){
                for(int j=0; j<nbColonneMatrice; j++){
                    DrawPoly((Vector2){matriceMines[i][j]->pos.x, matriceMines[i][j]->pos.y}, 6, SIZE_OF_HEX, 0, GRAY);
                    DrawPoly((Vector2){matriceMines[i][j]->pos.x, matriceMines[i][j]->pos.y}, 6, SIZE_OF_HEX-3, 0, Fade(matriceMines[i][j]->couleurCase, matriceMines[i][j]->isNeareast ? 0.5f : 1.0f));
                    
                    if(matriceMines[i][j]->isPressed && matriceMines[i][j]->nearbyMine>0) 
                        DrawText(TextFormat("%i", matriceMines[i][j]->nearbyMine), matriceMines[i][j]->pos.x - SIZE_OF_HEX/3, matriceMines[i][j]->pos.y - SIZE_OF_HEX/3, SIZE_OF_HEX*0.8, LIME);
                }
            }

        EndDrawing();
        //----------------------------------------------------------------------------------
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}